﻿using Eingabeverwaltung;
using Eingabeverwaltung.Metamodell;
using System;
using System.Collections.Generic;

namespace Ueberpruefung.CheckInduction
{
    /// <summary>
    /// checks error 38, 39
    /// </summary>
    public class CheckIndTask_Satisfied : ElementChecker
    {
        public CheckIndTask_Satisfied() : base() { }

        public override bool check(Element e)
        {
            if (e is Container con)
                if (con.Task.task.indVariable == null) return true;
                else return check(con.Task.task.indVariable, con.Proof.proof);
            else if (e is Lemma l)
                if (l.OwnTask != null && l.OwnTask.indVariable != null)
                {   // due to variables of lemmata can be renamed, this must also be considered at induction variable
                    CheckHelper.getSubsAndRenamings(l.eq, l.OwnTask.eq, out _, out List<Substitution> renamings);
                    Substitution indVarRenaming = FindVariableOfSubstitution(renamings, new TreeVariable(l.OwnTask.indVariable));
                    if (indVarRenaming != null) return check(indVarRenaming, l.proof);
                    else return check(l.OwnTask.indVariable, l.proof);
                }
                else return true;
            else throw new InvalidOperationException();
        }

        private bool check(Variable taskIndVar, SingleProof proof) => check(new Substitution(taskIndVar, new TreeVariable(taskIndVar)), proof);

        /// <param name="taskRenaming">Renaming of induction variable of proof to induction variable of Task.</param>
        /// <param name="proof"></param>
        /// <returns></returns>
        private bool check(Substitution taskRenaming, SingleProof proof)
        {
            bool result = true;
            Variable taskIndVar = ((TreeVariable)taskRenaming.tree).variable;
            if (proof is Transformation tr)
            { result = false; tr.errors.Add(new Error(39)); }
            else if (proof is Induction ind)
            {
                if (ind.indVar != taskRenaming.variable)
                { result = false; ind.errors.Add(new Error(38)); }
            }
            else throw new KontextException(proof.coords,"Found a single proof that is neither Transformation nor Induction!");
            return result;
        }

        /// <summary>
        /// Finds the first substitution in the list of substitutions where the tree is equal to the given tree.
        /// </summary>
        /// <param name="subs"></param>
        /// <param name="tree"></param>
        /// <returns></returns>
        private Substitution FindVariableOfSubstitution(List<Substitution> subs, Tree tree)
        {
            Substitution result = null;
            foreach(Substitution sub in subs)
                if (sub.tree.Equals(tree)) return sub;
            return result;
        }
    }
}
